/*
 *  copyright (c) 2006 IBM Corporation
 *  Authored by: Mike D. Day <ncmike@us.ibm.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/kobject.h>
#include <linux/sysfs.h>
#include <linux/err.h>
#include <asm/xen/hypervisor.h>
#include <xen/xenbus.h>
#include "xenctrl.h"

#define HYPERVISOR_ATTR_RO(_name) \
static struct kobj_attribute _name##_attr = __ATTR_RO(_name)

#define HYPERVISOR_ATTR_RW(_name) \
static struct kobj_attribute _name##_attr = \
	__ATTR(_name, 0644, _name##_show, _name##_store)

static ssize_t type_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	return sprintf(buffer, "xen\n");
}

HYPERVISOR_ATTR_RO(type);

static int __init xen_sysfs_type_init(void)
{
	return sysfs_create_file(hypervisor_kobj, &type_attr.attr);
}

static void xen_sysfs_type_destroy(void)
{
	sysfs_remove_file(hypervisor_kobj, &type_attr.attr);
}

static ssize_t major_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	int version;

	version = HYPERVISOR_xen_version(XENVER_version, NULL);
	if (!version)
		return -ENODEV;

	return sprintf(buffer, "%d\n", version >> 16);
}

HYPERVISOR_ATTR_RO(major);

static ssize_t minor_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	int version;

	version = HYPERVISOR_xen_version(XENVER_version, NULL);
	if (!version)
		return -ENODEV;

	return sprintf(buffer, "%d\n", version & 0xff);
}

HYPERVISOR_ATTR_RO(minor);

static ssize_t extra_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	int ret;
	struct xen_extraversion extra;

	ret = HYPERVISOR_xen_version(XENVER_extraversion, &extra);
	if (ret)
		return ret;

	return sprintf(buffer, "%s\n", extra.extraversion);
}

HYPERVISOR_ATTR_RO(extra);

static struct attribute *version_attrs[] = {
	&major_attr.attr,
	&minor_attr.attr,
	&extra_attr.attr,
	NULL
};

static struct attribute_group version_group = {
	.name = "version",
	.attrs = version_attrs,
};

static int __init xen_sysfs_version_init(void)
{
	return sysfs_create_group(hypervisor_kobj, &version_group);
}

static void xen_sysfs_version_destroy(void)
{
	sysfs_remove_group(hypervisor_kobj, &version_group);
}

static ssize_t uuid_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	char *vm, *val;
	int ret;

	vm = xenbus_read(XBT_NIL, "vm", "", NULL);
	if (IS_ERR(vm))
		return PTR_ERR(vm);

	val = xenbus_read(XBT_NIL, vm, "uuid", NULL);
	if (IS_ERR(val)) {
		ret = PTR_ERR(val);
		goto out;
	}

	ret = sprintf(buffer, "%s\n", val);

	kfree(val);
out:	kfree(vm);

	return ret;
}

HYPERVISOR_ATTR_RO(uuid);

static int __init xen_sysfs_uuid_init(void)
{
	return sysfs_create_file(hypervisor_kobj, &uuid_attr.attr);
}

static void xen_sysfs_uuid_destroy(void)
{
	sysfs_remove_file(hypervisor_kobj, &uuid_attr.attr);
}

static ssize_t compiler_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	struct xen_compile_info info;
	int ret;

	ret = HYPERVISOR_xen_version(XENVER_compile_info, &info);
	if (ret)
		return ret;

	return sprintf(buffer, "%s\n", info.compiler);
}

HYPERVISOR_ATTR_RO(compiler);

static ssize_t compiled_by_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	struct xen_compile_info info;
	int ret;

	ret = HYPERVISOR_xen_version(XENVER_compile_info, &info);
	if (ret)
		return ret;

	return sprintf(buffer, "%s\n", info.compile_by);
}

HYPERVISOR_ATTR_RO(compiled_by);

static ssize_t compile_date_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	struct xen_compile_info info;
	int ret;

	ret = HYPERVISOR_xen_version(XENVER_compile_info, &info);
	if (ret)
		return ret;

	return sprintf(buffer, "%s\n", info.compile_date);
}

HYPERVISOR_ATTR_RO(compile_date);

static struct attribute *xen_compile_attrs[] = {
	&compiler_attr.attr,
	&compiled_by_attr.attr,
	&compile_date_attr.attr,
	NULL
};

static struct attribute_group xen_compilation_group = {
	.name = "compilation",
	.attrs = xen_compile_attrs,
};

static int __init xen_compilation_init(void)
{
	return sysfs_create_group(hypervisor_kobj,
				  &xen_compilation_group);
}

static void xen_compilation_destroy(void)
{
	sysfs_remove_group(hypervisor_kobj,
			   &xen_compilation_group);
}

static ssize_t capabilities_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	struct xen_capabilities_info *caps;
	int ret;

	caps = kmalloc(sizeof(struct xen_capabilities_info), GFP_KERNEL);
	if (!caps)
		return -ENOMEM;

	ret = HYPERVISOR_xen_version(XENVER_capabilities, caps);
	if (ret)
		goto out;

	ret = sprintf(buffer, "%s\n", caps->info);

out:	kfree(caps);

	return ret;
}

HYPERVISOR_ATTR_RO(capabilities);

static ssize_t changeset_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	struct xen_changeset_info cset;
	int ret;

	ret = HYPERVISOR_xen_version(XENVER_changeset, &cset);
	if (ret)
		return ret;

	return sprintf(buffer, "%s\n", cset.info);
}

HYPERVISOR_ATTR_RO(changeset);

static ssize_t virtual_start_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	struct xen_platform_parameters parms;
	int ret;

	ret = HYPERVISOR_xen_version(XENVER_platform_parameters, &parms);
	if (ret)
		return ret;

	return sprintf(buffer, "%lx\n", parms.virt_start);
}

HYPERVISOR_ATTR_RO(virtual_start);

static ssize_t pagesize_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	int ret;

	ret = HYPERVISOR_xen_version(XENVER_pagesize, NULL);
	if (ret < 0)
		return ret;

	return sprintf(buffer, "%x\n", ret);
}

HYPERVISOR_ATTR_RO(pagesize);

static ssize_t writable_pt_show(struct kobject *kobj, struct kobj_attribute *attr, char *buffer)
{
	struct xen_feature_info info;
	int ret;

	info.submap_idx = XENFEAT_writable_page_tables;

	ret = HYPERVISOR_xen_version(XENVER_get_features, &info);
	if (ret)
		return ret;

	return sprintf(buffer, "%d\n", info.submap);
}

HYPERVISOR_ATTR_RO(writable_pt);

static struct attribute *xen_properties_attrs[] = {
	&capabilities_attr.attr,
	&changeset_attr.attr,
	&virtual_start_attr.attr,
	&pagesize_attr.attr,
	&writable_pt_attr.attr,
	NULL
};

static struct attribute_group xen_properties_group = {
	.name = "properties",
	.attrs = xen_properties_attrs,
};

static int __init xen_properties_init(void)
{
	return sysfs_create_group(hypervisor_kobj,
				  &xen_properties_group);
}

static void xen_properties_destroy(void)
{
	sysfs_remove_group(hypervisor_kobj, &xen_properties_group);
}

int __init sys_hypervisor_init(void)
{
	int ret;

	if (!is_running_on_xen())
		return -ENODEV;

	ret = xen_sysfs_type_init();
	if (ret)
		goto out;
	ret = xen_sysfs_version_init();
	if (ret)
		goto version_out;
	ret = xen_compilation_init();
	if (ret)
		goto comp_out;
	ret = xen_sysfs_uuid_init();
	if (ret)
		goto uuid_out;
	ret = xen_properties_init();
	if (!ret)
		goto out;

	xen_sysfs_uuid_destroy();
uuid_out:
	xen_compilation_destroy();
comp_out:
	xen_sysfs_version_destroy();
version_out:
	xen_sysfs_type_destroy();
out:
	return ret;
}

void __exit sys_hypervisor_exit(void)
{
	xen_properties_destroy();
	xen_compilation_destroy();
	xen_sysfs_uuid_destroy();
	xen_sysfs_version_destroy();
	xen_sysfs_type_destroy();
}
